"""

NOTE: This code runs if kawaplatformlibrary is already installed in the virtual environment. 

"""

"""
Example for downloading one Sentinel 2 tile for a specific AOI. All parameters are optional and contain default values which will be used for downloading the Sentinel 2 tile. The only parameter which is required is the destination file for storing the TIF obtained from STAC API.s

Input parameters:
    1. geojson                  => GeoJSON for the Area of Interest[AOI]. Optional with a default value
    2. start_date               => The initial date from when data should be requested from STAC API. Format maintained is YYYY-MM-DD. Optional with a default value.
    3. end_date                 => The final date before which data should be requested from STAC API. Format maintained is YYYY-MM-DD. Optional with a default value.
    4. bands                    => The bands to be requested from the STAC API for the Sentinel 2 tiles. Comma separated list. Optional with default being ["B02","B03","B04"]
    5. ground_sampling_distance => The ground sampling distance to which all bands should be resampled to. Optional with a default value of 10m.
    6. destination_file         => The file location where the obtained TIF file should be stored.

Output parameters:
    Stores the data obtained in a file specified by the user.

Example UseCase : 
python download_tiles.py --geojson ../../test_aoi.geojson --start_date 2020-02-01 --end_date 2020-02-27 --bands B02,B03,B04 --ground_sampling_distance 10 --destination_folder ../../Test_Examples_Folder_Path --num_threads 6
"""

# importing libraries
import numpy as np
import json, argparse, rasterio, os

import sys
sys.path.insert(1, "./../") # Changing the system path to root directory of the library. This is needed to actually call the Kawa Platform library

from kawaplatformlibrary.data import sentinel2dataingestion
from kawaplatformlibrary.data.preprocessing import splitgeojson
from kawaplatformlibrary.postprocessing.mosiac import mergeBands
from kawaplatformlibrary.postprocessing.createprofile import createNewProfile
from kawaplatformlibrary.postprocessing.creatememoryraster import createMemRaster
from kawaplatformlibrary.postprocessing.extractarea import (reprojectPolygon, extractArea)

from rasterio import Affine

# Optional arguments for the user
ap = argparse.ArgumentParser()
ap.add_argument("-g", "--geojson", required=True, 
    help="Path to GEOJSON file containing the AOI.")
ap.add_argument("-s", "--start_date", required=True, 
    help="Start date in YYYY-MM-DD format")
ap.add_argument("-e", "--end_date", required=True, 
    help="End date in YYYY-MM-DD format")
ap.add_argument("-b", "--bands", required=True,
    help="Comma separated list of bands")
ap.add_argument("-d","--ground_sampling_distance", nargs="?", default=10, type=int,
    help="Ground Sampling Distance required for each band")
# This is not optional. User should input this
ap.add_argument("-f", "--destination_folder", required=True,
    help="Path to folder for storage. Use the full OS path and not relative path")
args = vars(ap.parse_args())

user_bands = [band.strip() for band in args["bands"].split(',')]

# Reading the GeoJSON of the AOI and extracting the coordinates json. {"type": "Polygon", "coordinates":[[[...], [...], ...]]]}
with open(args["geojson"], "r") as in_file:
    geojson_contents = json.load(in_file)
    geojson_coordinates = geojson_contents["features"][0]["geometry"]
    in_file.close()
    pass

# Obtaining image url from STAC API. dataCheck = [0/1, "[ERROR]/["INFO"]", {}/{<data>}]
print("[INFO] Obtianing data for the AOI")
sentinel2_data_class_obtain_data = sentinel2dataingestion.ObtainData(bands=user_bands)
sentinel2_data_aoi = sentinel2_data_class_obtain_data.getBandData(geojson=geojson_coordinates, start_date=args["start_date"], end_date=args["end_date"])
print("[INFO] Finished finding data for the AOI.")

def downloadData(rasters_href):
    """
    Downloading and storing the data in the destination folder.
    """
    print("[INFO] Downloading data for {}".format(rasters_href["img_id"]))

    destination_file = args["destination_folder"] + rasters_href["img_id"] + ".tif"

    bands_rasters_list = []
    bands_data = rasters_href["band_data"]

    # Reading the bands into memory as numpy arrays
    for j, band in enumerate(bands_data):

        band_href = bands_data[band]["href"]
        band_src = rasterio.open(band_href)
        bands_rasters_list.append(band_src)

        # Creating the destination profile for each Sentinel 2 tile. Will be used for storing the current tile.
        if j == 0:
            num_bands = len(user_bands)
            dst_profile = createNewProfile(band_src.transform, band_src.profile, ground_sampling_distance=10, num_bands=num_bands)
            pass
        pass

    # Resampling and merging the bands
    print("[INFO] Merging bands for img : {}".format(rasters_href["img_id"]))
    merged_resample_bands = mergeBands(bands_rasters_list, args["ground_sampling_distance"])
    print("[INFO] Finished merging bands for img : {}".format(rasters_href["img_id"]))

    print("[INFO] Creating merged bands raster.")
    merged_bands_raster = createMemRaster(dst_profile, merged_resample_bands)
    print("[INFO] Finished creating raster.")

    print("[INFO] Reprojecting polygon to required CRS.")
    bands_epsg = rasters_href["init"]
    geojson_coordinates_reproject = reprojectPolygon(gj=geojson_coordinates, outproj=bands_epsg, inproj="EPSG:4326")
    print("[INFO] Finished reprojecting polygon to required CRS.")

    print("[INFO] Clipping area.")
    geojson_clipped_area, geojson_clipped_area_profile = extractArea(geojson_coordinates_reproject, merged_bands_raster)
    print("[INFO] Finished clipping area.")

    print("[INFO] Storing clipping data.")
    # Storing the Sentinel 2 tile in the destination folder with tile name corresponding to the tile ID
    with rasterio.open(destination_file, "w", **geojson_clipped_area_profile) as out_file:
        out_file.write(geojson_clipped_area)
        pass

    print("[INFO] Finished downloading data for {}".format(rasters_href["img_id"]))
    pass

data = list(sentinel2_data_aoi.items())[0][1]
data_sensing_date = list(sentinel2_data_aoi.items())[0][0]
data_check = data[0]

if data_check == 1:
    downloadData(data[2])
else:
    print("[INFO] Error Code : {} Error Message : {}".format(data[0], data[1]))


"""
Calculation of Global Environment Monitoring Index using the Near-Infrared and Red bands

Formula to calculate variable 'n': ( 2 * ( NIR ^2 - RED ^2) + 1.5 * NIR + 0.5 * RED ) / ( NIR + RED + 0.5 )

Formula to calculate GEMI: ( n * (1- 0.25n) ) - ( (RED - 0.125) / (1-RED) )

Wavelengths: 640nm :760nm , 780nm : 1400nm
"""

import numpy as np
import pandas as pd
import rasterio, os

from kawaplatformlibrary.postprocessing.mosiac import reSampleBand

def store():
    """
    This will be done when we have decided on the data storage strategy.
    """
    pass

def calculate(band_red, band_nir, ground_sampling_distance):
    """
    Function for calculating the GEMI
    
    Input Parameters:
        1. band_nir => NIR band of the satellite image
        2. band_red => Red band of the satellite image

    Output Parameters:
        1. gemi => GEMI for the entire satellite image
    """

    if band_nir.transform[0] != ground_sampling_distance:
        band_nir_array = reSampleBand(band_nir, ground_sampling_distance)
    else:
        band_nir_array = np.array(band_nir.read(1), dtype=np.float32)

    if band_red.transform[0] != ground_sampling_distance:
        band_red_array = reSampleBand(band_red, ground_sampling_distance)
    else:
        band_red_array = np.array(band_red.read(1), dtype=np.float32)
    
    n_var = (2*(band_nir_array**2 - band_red_array**2) + 1.5*band_nir_array + 0.5*band_red_array)/(band_nir_array + band_red_array + 0.5) 
    gemi = (n_var*(1-(0.25*n_var))) - ((band_red_array-0.125)/(1-band_red_array))

    return np.array(gemi, dtype=np.float32)
    pass



__version__ = "V1.0"
__contributors__ = "Gotam Dahiya, Manya Chadha"
__documentation__ = "https://github.com/KawaSpaceOrg/kawaplatformlibrary/wiki"
__homepage__ = "https://github.com/KawaSpaceOrg/kawaplatformlibrary" 


"""
Calculation of Normalised Difference Red-edge using the Near-Infrared and Red bands

Formula to calculate NDRE: (NIR - RED_edge) / (NIR + RED_edge)

Wavelengths: 690nm :730nm , 780nm : 1400nm
"""

import numpy as np
import pandas as pd
import rasterio, os

from kawaplatformlibrary.postprocessing.mosiac import reSampleBand

def store():
    """
    This will be done when we have decided on the data storage strategy.
    """
    pass

def calculate(band_red_edge, band_nir, ground_sampling_distance):
    """
    Function for calculating the NDRE
    
    Input Parameters:
        1. band_nir => NIR band of the satellite image
        2. band_red => Red band of the satellite image

    Output Parameters:
        1. ndre => NDRE for the entire satellite image
    """

    if band_nir.transform[0] != ground_sampling_distance:
        band_nir_array = reSampleBand(band_nir, ground_sampling_distance)
    else:
        band_nir_array = np.array(band_nir.read(1), dtype=np.float32)

    if band_red_edge.transoform[0] != ground_sampling_distance:
        band_red_edge_array = reSampleBand(band_red_edge, ground_sampling_distance)
    else:
        band_red_edge_array = np.array(band_red_edge.read(1), dtype=np.float32)
    
    ndre = (band_nir_array - band_red_edge_array)/(band_nir_array + band_red_edge_array)

    return np.array(ndre, dtype=np.float32)
    pass



"""
Calculation of Modified Triangular Vegetation Index 1[MTVI_1] using the 800nm, 550nm and 670nm bands

Formula : 1.2 * (1.2 * (800nm - 550nm) - 2.5 * (670nm - 550nm))
Wavelengths : 550nm, 670nm, 800nm
"""

import numpy as np
import pandas as pd
import rasterio, os

from kawaplatformlibrary.postprocessing.mosiac import reSampleBand

def store():
    """
    This will be done when we have decided on the data storage strategy.
    """
    pass

def calculate(band_550, band_670, band_800, ground_sampling_distance):
    """
    Function for calculating the CIGreen
    
    Input Parameters:
        1. band_550 => 550nm band of the satellite image
        2. band_670 => 670nm band of the satellite image
        3. band_800 => 800nm band of the satellite image

    Output Parameters:
        1. mtvi_one => MTVI_1 for the entire satellite image
    """

    if band_550.transform[0] != ground_sampling_distance:
        band_550_array = np.where(band_550_array == band_550.meta["nodata"], np.NaN, band_550_array)
    else:
        band_550_array = np.array(band_550,read(1), dtype=np.float32)

    if band_670.transform[0] != ground_sampling_distance:
        band_670_array = reSampleBand(band_670, ground_sampling_distance)
    else:
        band_670_array = np.array(band_670.read(1), dtype=np.float32)

    if band_800.transform[0] != ground_sampling_distance:
        band_800_array = reSampleBand(band_800, ground_sampling_distance)
    else:
        band_800_array = np.array(band_800.read(1), dtype=np.float32)

    mtvi_one = 1.2 * (1.2*(band_800_array - band_550_array) - 2.5*(band_670_array - band_550_array))

    return np.array(mtvi_one, dtype=np.float32)
    pass



"""
Calculation of Enhanced Vegetation Index[EVI] using the NIR, Red and Blue bands

Formula : 2.5 * ((NIR - RED) / (NIR + 6*RED - 7.5*BLUE + 1))
Wavelengths : NIR, Red and Blue
"""

import numpy as np
import pandas as pd
import rasterio, os

from kawaplatformlibrary.postprocessing.mosiac import reSampleBand

def store():
    """
    This will be done when we have decided on the data storage strategy.
    """
    pass

def calculate(band_blue, band_red, band_nir, ground_sampling_distance):
    """
    Function for calculating the CIGreen
    
    Input Parameters:
        1. band_nir  => NIR band of the satellite image
        2. band_red  => Red band of the satellite image
        3. band_blue => Blue band of the satellite image

    Output Parameters:
        1. evi => EVI for the entire satellite image
    """

    if band_nir.transform[0] != ground_sampling_distace:
        band_nir_array = reSampleBand(band_nir, ground_sampling_distance)
    else:
        band_nir_array = np.array(band_nir,read(1), dtype=np.float32)

    if band_red.transform[0] != ground_sampling_distance:
        band_red_array = reSampleBand(band_red, ground_sampling_distance)
    else:
        band_red_array = np.array(band_red.read(1), dtype=np.float32)

    if band_blue.transform[0] != ground_sampling_distance:
        band_blue_array = reSampleBand(band_blue, ground_sampling_distance)
    else:
        band_blue_array = np.array(band_blue.read(1), dtype=np.float32)
    
    evi = 2.5 * ((band_nir_array - band_red_array) / (band_nir_array + 6*band_red_array - 7.5*band_blue_array + 1))

    return np.array(evi, dtype=np.float32)
    pass



import numpy as np
import json, argparse, datetime

from shapely.geometry import Point, Polygon, box, mapping, shape, LineString
from sklearn.metrics import pairwise_distances_argmin_min

from pathlib import Path

def degreeToKm(degree):
    return degree*111

def kmToDegree(distance_km):
    return distance_km/111


def getSatelliteCentroids(satellite_system):
    """
    Obtaining the geojson containing the satellite data polygon centroids

    Input Parameters :
        1. satellite_system => The satellite systems data is being obtained from

    Output Parameters : 
        1. List of coordinates of the centroids of the satellite data polygons
    """

    current_file_path = Path(__file__).parent.absolute()

    satellite_dict = {"SENTINEL2" : current_file_path/"satellitecentroids/sentinel2_centroid_geojsons.geojson"}

    with open(satellite_dict[satellite_system], "r") as in_file:
        geojson_contents = json.load(in_file)
        geojson_coordinates = geojson_contents["features"]
        in_file.close()
        pass

    return geojson_coordinates
    pass

def getClosestSatelliteCentroid(satellite_system, split_centroid_coords):
    """
    Obtaining the closest satellite polygon centroid to current point depending upon the satellite system

    Input Parameters :
        1. satellite_system      => Which satellite system is being used
        2. split_centroid_coords => The point to which the closest satellite polygon centroid has to be obtained

    Output Parameters :
        1. Returns the closest satellite polygon point
    """
    
    split_centroid_coords = list(*zip(*split_centroid_coords.coords.xy))

    satellite_centroid_geojson_features = getSatelliteCentroids(satellite_system)
    def get_centroid_coords():
        satellite_centroid_coordinates_list = []

        for coords in satellite_centroid_geojson_features:

            satellite_centroid_long = coords["geometry"]["coordinates"][0]
            satellite_centroid_lat  = coords["geometry"]["coordinates"][1]

            satellite_centroid_coordinates_list.append([satellite_centroid_long, satellite_centroid_lat])  
            pass

        return np.array(satellite_centroid_coordinates_list, dtype=np.float32)
        pass


    satellite_centroid_list = get_centroid_coords()
    split_geojson_coordinates = [[split_centroid_coords[0], split_centroid_coords[1]]]

    closest, _ = pairwise_distances_argmin_min(split_geojson_coordinates, satellite_centroid_list)

    return Point(satellite_centroid_list[closest][0])
    pass

def split(satellite_system, geojson_file, side_length_km=100):
    
    '''
    Splits a larger AOI into smaller AOIs of a desired side length.

    Input Parameters
    satellite_system => From which satellite system the data is being sourced
    geojson_file     => Coordinates of the AOI
    side_length_km   => side length of the square/tile desired

    Output Parameters
    '''


    user_poly = Polygon(geojson_file["coordinates"][0])
    usr_in = user_poly.buffer(kmToDegree(side_length_km/2),join_style=2)
    
    '''create a buffer around user input polygon and then create envelope
    start from one end of the box, find diagonally opposite point of side_lengthxside_length square
    draw the diagonal and find the centroid
    create buffer around centroid
    find next centroid within bounding box by moving side_length units along x axis'''
    
    env = usr_in.envelope # bounding box
    # print(env)
    start = Point(mapping(env)['coordinates'][0][0])  # startpoint

    p = Point(start.x + kmToDegree(side_length_km),start.y + kmToDegree(side_length_km)) # diagonally opposite point 
    cen_next_x = LineString([start, p]).centroid  # centroid along diagonal 

    final_geojsons = []

    half_side = side_length_km/2 # Size for calculating the buffer

    tenth_side = side_length_km/100


    # Finding the smaller polygons within the bounding box of the original AOI. 
    # First while loop is for the first row starting the from the bottom left corner
    # The second while loop is for the column starting with the current first row polygon. 
    while True:
        
        cen_satellite_next_x = getClosestSatelliteCentroid(satellite_system, cen_next_x) # Obtaining the closest satellite centroid

        buf_x = mapping(cen_satellite_next_x.buffer(kmToDegree(tenth_side),cap_style=3)) # buffer around centroid

        # Storing the polygon in final_geojsons, if the polygon intersects with the bounding box
        buf_shp_x = Polygon(buf_x["coordinates"][0])
        if usr_in.intersects(buf_shp_x):
            final_geojsons.append(buf_x)
            pass

        cen_next_y = cen_next_x # Making the initial y centroid the same as the x centroid 

        while True:
            cen_next_y = Point(cen_next_y.x + kmToDegree(0), cen_next_y.y + kmToDegree(side_length_km))

            cen_satellite_next_y = getClosestSatelliteCentroid(satellite_system, cen_next_y) # Obtaining the closest satellite centroid

            if not cen_next_y.within(env): # If the centroid calculated is beyond the bounding box, break the current loop
                break
                pass

            buf_y = mapping(cen_satellite_next_y.buffer(kmToDegree(tenth_side),cap_style=3))

            # Storing the polygon in final_geojsons, if the polygon intersects with the bounding box
            buf_shp_y = Polygon(buf_y["coordinates"][0])
            if usr_in.intersects(buf_shp_y):
                final_geojsons.append(buf_y)
                pass
            pass

        cen_next_x = Point(cen_next_x.x + kmToDegree(side_length_km),cen_next_x.y + kmToDegree(0))

        if not cen_next_x.within(env): # If the centroid calculated is beyond the bounding box, break the current loop
            break
            pass
        pass
            

    return final_geojsons
    # return centroids #for testing new split method 
    pass


from shapely.geometry import shape, Polygon, Point

def createBBox(pos, side_km):

    side_km = 0.009120 * side_km
    min_lat = pos[1] - side_km
    max_lat = pos[1] + side_km
    min_lon = pos[0] - side_km
    max_lon = pos[0] + side_km

    lower_left  = Point(min_lon, min_lat)
    lower_right = Point(max_lon, min_lat)
    upper_right = Point(max_lon, max_lat)
    upper_left  = Point(min_lon, max_lat)

    point_list = [lower_left, lower_right, upper_right, upper_left]

    poly = Polygon([[p.x, p.y] for p in point_list])

    polyg = poly.exterior.coords

    coords = []
    for x,y in polyg:
        coords.append([x,y])
        pass

    gj = {"type" : "Polygon", "coordinates" : [coords]}

    return gj
    pass


"""
Calculation of Normalised Difference Water Index (Crop Water) using Near Infrared and Short Wave Infrared bands

Formula : (NIR - SWIR) / (NIR + SWIR)

Wavelengths : 780nm :1400nm , 1400nm :3000nm
"""

import numpy as np
import pandas as pd
import rasterio, os

from kawaplatformlibrary.postprocessing.mosiac import reSampleBand

def store():
    """
    This will be done when we have decided on the data storage strategy.
    """
    pass

def calculate(band_nir, band_swir, ground_sampling_distance):
    """
    Function for calculating the NDWI (crop water)
    
    Input Parameters:
        1. band_nir => NIR band of the satellite image
        2. band_swir => SWIR band of the satellite image

    Output Parameters:
        1. ndwi_cw => NDWI (crop water) for the entire satellite image
    """

    if band_nir.transform[0] != ground_sampling_distance:
        band_nir_array = reSampleBand(band_nir, ground_sampling_distance)
    else:
        band_nir_array = np.array(band_nir.read(1), dtype=np.float32)

    if band_swir.transform[0] != ground_sampling_distance:
        band_swir_array = reSampleBand(band_swir, ground_sampling_distance)
    else:
        band_swir_array = np.array(band_swir.read(1), dtype=np.float32)


    ndwi_cw = (band_nir_array - band_swir_array)/(band_nir_array + band_swir_array)

    return np.array(ndwi_cw, dtype=np.float32)
    pass



"""
To create a new profile for any raster. This does not create a new raster instead creates a profile for the new raster so that it can be written to a GeoTIFF file.
"""

import json
import rasterio

from rasterio import Affine

def createNewProfile(src_transform, src_profile, ground_sampling_distance=None, num_bands=1, dtype="float32", img_height=None, img_width=None):
    """
    Creating a function for making a new Profile with change in the Ground Sampling distance and Count of bands

    Input Parameters:
        1. src_transform             => Transformation matrix of the source raster
        2. src_profile               => Profile dictionary of the source raster
        3. ground_sampling_distance  => Ground sampling distance of the resultant profile
        4. num_bands                 => Number of bands present in the resultant profile
        5. dtype                     => Decimal type of the resultant profile. Default is set to float32.
        6. img_height                => Specify the height of the image to which the profile is to fit. Defaults to None
        7. img_width                 => Specify the width of the image to which the profile is to fit. Defaults to None

        img_height and img_width both have to be included. They do not work separately.

    Output:
        1. New and updated profile.
    """
    if ground_sampling_distance and src_transform[0] != ground_sampling_distance:
        scale = src_transform[0] // ground_sampling_distance
    else:
        scale = 1

    dst_transform = Affine(src_transform.a / scale, src_transform.b, src_transform.c, src_transform.d, 
        src_transform.e / scale, src_transform.f)

    if img_height and img_width:
        dst_height = img_height
        dst_width  = img_width
    else: 
        dst_width = int(src_profile["height"] * scale)
        dst_height = int(src_profile["width"] * scale)

    dst_profile = src_profile
    dst_profile.update(transform=dst_transform, driver="GTIFF", height=dst_height, width=dst_width, 
        count=num_bands, dtype="float32")

    return dst_profile
    pass


"""
Calculation of Browning Reflectance Index using the 550nm, 700nm and NIR bands

Formula: ((1/550 nm) - (1/700 nm)) / (NIR)

Wavelengths: 550nm ,700nm ,780nm :1400nm
"""

import numpy as np
import pandas as pd
import rasterio, argparse, os

from kawaplatformlibrary.postprocessing.mosiac import reSampleBand

def store():
    """
    This will be done when we have decided on the data storage strategy.
    """
    pass

def calculate(band_550, band_700, band_nir, ground_sampling_distance):
    """
    Function for calculating the BRI
    
    Input Parameters:
        1. band_nir => NIR band of the satellite image
        2. band_550 => 550nm spectral band of the satellite image
        3. band_700 => 700nm spectral band of the satellite image
    Output Parameters:
        1. bri => BRI for the entire satellite image
    """


    if band_nir.transform[0] != ground_sampling_distance:
        band_nir_array = reSampleBand(band_nir, ground_sampling_distance)
    else:
        band_nir_array = np.array(band_nir.read(1), dtype=np.float32)

    if band_550.transform[0] != ground_sampling_distance:
        band_550_array = reSampleBand(band_550, ground_sampling_distance)
    else:
        band_550_array = np.array(band_550.read(1), dtype=np.float32)

    if band_700.transform[0] != ground_sampling_distance:
        band_700_array = reSampleBand(band_700, ground_sampling_distance)
    else:
        band_700_array = np.array(band_700.read(1), dtype=np.float32)

    bri = ((1/band_550_array)-(1/band_700_array))/band_nir_array

    return np.array(bri, dtype=np.float32)
    pass



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE/"PYPI_README.md").read_text()

setup(
    name="kawaplatformlibrary",
    version="1.0.3",
    description="Library for Data Ingestion",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/KawaSpaceOrg/kawaplatformlibrary",
    license_files = ("LICENSE.md"),
    author="Gotam",
    author_email="gotamdahiya@gmail.com",
    packages=find_packages(exclude=("bin", "lib", "examples", "test" )),
    include_package_data=True,
    install_requires=["joblib>=1.2.0", "numpy>=1.23.5", "pandas>=1.5.2", "python_dateutil>=2.8.2", "rasterio>=1.3.4", "requests>=2.28.1", "scikit_learn>=1.1.3", "Shapely>=18.5.post1", "pyproj>=3.4.0"],
    python_requires='>=3.8'
    )